@@ -25,6 +25,7 @@
 #include "scram.h"
 
 #include <usual/pgutil.h>
+#include <usual/slab.h>
 
 static const char *hdr2hex(const struct MBuf *data, char *buf, unsigned buflen)
 {
@@ -52,7 +53,7 @@ PgDatabase *prepare_auth_database(PgSocket *client)
 	if (!auth_dbname) {
 		auth_db = client->db;
 	} else {
-		auth_db = find_database(auth_dbname);
+		auth_db = find_or_register_database(client, auth_dbname);
 	}
 
 	if (!auth_db) {
@@ -98,7 +99,8 @@ static bool check_client_passwd(PgSocket *client, const char *passwd)
 			return strcmp(user->passwd, passwd) == 0;
 		case PASSWORD_TYPE_MD5: {
 			char md5[MD5_PASSWD_LEN + 1];
-			pg_md5_encrypt(passwd, user->name, strlen(user->name), md5);
+			if (!pg_md5_encrypt(passwd, user->name, strlen(user->name), md5))
+				return false;
 			return strcmp(user->passwd, md5) == 0;
 		}
 		case PASSWORD_TYPE_SCRAM_SHA_256:
@@ -121,12 +123,14 @@ static bool check_client_passwd(PgSocket *client, const char *passwd)
 		 * md5() call first.
 		 */
 		if (get_password_type(user->passwd) == PASSWORD_TYPE_PLAINTEXT) {
-			pg_md5_encrypt(user->passwd, user->name, strlen(user->name), md5);
+			if (!pg_md5_encrypt(user->passwd, user->name, strlen(user->name), md5))
+				return false;
 			stored_passwd = md5;
 		} else {
 			stored_passwd = user->passwd;
 		}
-		pg_md5_encrypt(stored_passwd + 3, (char *)client->tmp_login_salt, 4, md5);
+		if (!pg_md5_encrypt(stored_passwd + 3, (char *)client->tmp_login_salt, 4, md5))
+			return false;
 		return strcmp(md5, passwd) == 0;
 	}
 	}
@@ -140,7 +144,7 @@ static bool send_client_authreq(PgSocket *client)
 
 	if (auth_type == AUTH_MD5) {
 		uint8_t saltlen = 4;
-		get_random_bytes((void*)client->tmp_login_salt, saltlen);
+		get_random_bytes((void *)client->tmp_login_salt, saltlen);
 		SEND_generic(res, client, 'R', "ib", AUTH_MD5, client->tmp_login_salt, saltlen);
 	} else if (auth_type == AUTH_PLAIN || auth_type == AUTH_PAM) {
 		SEND_generic(res, client, 'R', "i", AUTH_PLAIN);
@@ -151,7 +155,7 @@ static bool send_client_authreq(PgSocket *client)
 	}
 
 	if (!res) {
-	    slog_noise(client, "No authentication response received");
+		slog_noise(client, "No authentication response received");
 		disconnect_client(client, false, "failed to send auth req");
 	} else {
 		slog_noise(client, "Auth request sent successfully");
@@ -292,8 +296,7 @@ static bool finish_set_pool(PgSocket *client, bool takeover)
 				client->db->name, client->login_user->name);
 	}
 
-	if (auth == AUTH_MD5)
-	{
+	if (auth == AUTH_MD5) {
 		if (get_password_type(client->login_user->passwd) == PASSWORD_TYPE_SCRAM_SHA_256)
 			auth = AUTH_SCRAM_SHA_256;
 	}
@@ -335,12 +338,7 @@ bool set_pool(PgSocket *client, const char *dbname, const char *username, const
 	Assert((password && takeover) || (!password && !takeover));
 
 	/* find database */
-	client->db = find_database(dbname);
-	if (!client->db) {
-		client->db = register_auto_database(dbname);
-		if (client->db)
-			slog_info(client, "registered new auto-database: db=%s", dbname);
-	}
+	client->db = find_or_register_database(client, dbname);
 	if (!client->db) {
 		client->db = calloc(1, sizeof(*client->db));
 		client->db->fake = true;
@@ -405,9 +403,9 @@ bool set_pool(PgSocket *client, const char *dbname, const char *username, const
 					client->db->auth_user = add_user(cf_auth_user, "");
 			}
 			if (client->db->auth_user) {
-				if (client->db->fake)
+				if (client->db->fake) {
 					slog_debug(client, "not running auth_query because database is fake");
-				else {
+				} else {
 					if (takeover) {
 						client->login_user = add_db_user(client->db, username, password);
 						return finish_set_pool(client, takeover);
@@ -427,14 +425,15 @@ bool set_pool(PgSocket *client, const char *dbname, const char *username, const
 	return finish_set_pool(client, takeover);
 }
 
-bool handle_auth_query_response(PgSocket *client, PktHdr *pkt) {
+bool handle_auth_query_response(PgSocket *client, PktHdr *pkt)
+{
 	uint16_t columns;
 	uint32_t length;
 	const char *username, *password;
 	PgUser user;
 	PgSocket *server = client->link;
 
-	switch(pkt->type) {
+	switch (pkt->type) {
 	case 'T':	/* RowDescription */
 		if (!mbuf_get_uint16be(&pkt->data, &columns)) {
 			disconnect_server(server, false, "bad packet");
@@ -487,7 +486,7 @@ bool handle_auth_query_response(PgSocket *client, PktHdr *pkt) {
 				return false;
 			}
 		}
-		if (sizeof(user.passwd)  - 1 < length)
+		if (sizeof(user.passwd) - 1 < length)
 			length = sizeof(user.passwd) - 1;
 		memcpy(user.passwd, password, length);
 
@@ -505,7 +504,7 @@ bool handle_auth_query_response(PgSocket *client, PktHdr *pkt) {
 		break;
 	case '2':	/* BindComplete */
 		break;
-	case 'S': /* ParameterStatus */
+	case 'S':	/* ParameterStatus */
 		break;
 	case 'Z':	/* ReadyForQuery */
 		sbuf_prepare_skip(&client->link->sbuf, pkt->len);
@@ -560,7 +559,7 @@ static bool read_escaped_token(const char **escaped_string_ptr, struct MBuf *une
 	const char *unwritten_start = position;
 	while (*position) {
 		if (*position == '\\') {
-			if (!mbuf_write(unescaped_token, unwritten_start, position-unwritten_start))
+			if (!mbuf_write(unescaped_token, unwritten_start, position - unwritten_start))
 				return false;
 			position++;
 			unwritten_start = position;
@@ -571,7 +570,7 @@ static bool read_escaped_token(const char **escaped_string_ptr, struct MBuf *une
 		}
 		position++;
 	}
-	if (!mbuf_write(unescaped_token, unwritten_start, position-unwritten_start))
+	if (!mbuf_write(unescaped_token, unwritten_start, position - unwritten_start))
 		return false;
 	if (!mbuf_write_byte(unescaped_token, '\0'))
 		return false;
@@ -848,7 +847,7 @@ static bool handle_client_startup(PgSocket *client, PktHdr *pkt)
 	}
 
 	if (client->wait_for_welcome || client->wait_for_auth) {
-		if  (finish_client_login(client)) {
+		if (finish_client_login(client)) {
 			/* the packet was already parsed */
 			sbuf_prepare_skip(sbuf, pkt->len);
 			return true;
@@ -966,8 +965,7 @@ static bool handle_client_startup(PgSocket *client, PktHdr *pkt)
 					free_scram_state(&client->scram_state);
 					if (!finish_client_login(client))
 						return false;
-				}
-				else {
+				} else {
 					disconnect_client(client, true, "SASL authentication failed");
 					return false;
 				}
@@ -1007,8 +1005,7 @@ static bool handle_client_startup(PgSocket *client, PktHdr *pkt)
 		break;
 	case PKT_CANCEL:
 		if (mbuf_avail_for_read(&pkt->data) == BACKENDKEY_LEN
-		    && mbuf_get_bytes(&pkt->data, BACKENDKEY_LEN, &key))
-		{
+		    && mbuf_get_bytes(&pkt->data, BACKENDKEY_LEN, &key)) {
 			memcpy(client->cancel_key, key, BACKENDKEY_LEN);
 			accept_cancel_request(client);
 		} else {
@@ -1029,9 +1026,12 @@ static bool handle_client_work(PgSocket *client, PktHdr *pkt)
 {
 	SBuf *sbuf = &client->sbuf;
 	int rfq_delta = 0;
+	int track_outstanding = false;
+	PreparedStatementAction ps_action = PS_IGNORE;
+	PgClosePacket close_packet;
+	PgPool *pool = client->pool;
 
 	switch (pkt->type) {
-
 	/* one-packet queries */
 	case 'Q':		/* Query */
 		if (cf_disable_pqexec) {
@@ -1040,14 +1040,17 @@ static bool handle_client_work(PgSocket *client, PktHdr *pkt)
 			return false;
 		}
 		rfq_delta++;
+		track_outstanding = true;
 		break;
 	case 'F':		/* FunctionCall */
 		rfq_delta++;
+		track_outstanding = true;
 		break;
 
 	/* request immediate response from server */
 	case 'S':		/* Sync */
 		rfq_delta++;
+		track_outstanding = true;
 		break;
 	case 'H':		/* Flush */
 		break;
@@ -1062,10 +1065,41 @@ static bool handle_client_work(PgSocket *client, PktHdr *pkt)
 	 * to buffer packets until sync or flush is sent by client
 	 */
 	case 'P':		/* Parse */
+		track_outstanding = true;
+		if (is_prepared_statements_enabled(pool)) {
+			ps_action = inspect_parse_packet(client, pkt);
+			pkt_rewind_v3(pkt);
+		}
+		break;
+
 	case 'E':		/* Execute */
+		track_outstanding = true;
+		break;
+
 	case 'C':		/* Close */
+		track_outstanding = true;
+		if (is_prepared_statements_enabled(pool)) {
+			ps_action = inspect_describe_or_close_packet(client, pkt);
+			pkt_rewind_v3(pkt);
+		}
+		break;
+
 	case 'B':		/* Bind */
+		track_outstanding = true;
+		if (is_prepared_statements_enabled(pool)) {
+			ps_action = inspect_bind_packet(client, pkt);
+			pkt_rewind_v3(pkt);
+		}
+		break;
+
 	case 'D':		/* Describe */
+		track_outstanding = true;
+		if (is_prepared_statements_enabled(pool)) {
+			ps_action = inspect_describe_or_close_packet(client, pkt);
+			pkt_rewind_v3(pkt);
+		}
+		break;
+
 	case 'd':		/* CopyData(F/B) */
 		break;
 
@@ -1074,11 +1108,87 @@ static bool handle_client_work(PgSocket *client, PktHdr *pkt)
 		slog_error(client, "unknown pkt from client: %u/0x%x", pkt->type, pkt->type);
 		disconnect_client(client, true, "unknown pkt");
 		return false;
-	case 'X': /* Terminate */
+	case 'X':	/* Terminate */
 		disconnect_client(client, false, "client close request");
 		return false;
 	}
 
+	if (ps_action == PS_HANDLE_FULL_PACKET && incomplete_pkt(pkt)) {
+		if (pkt->len > (unsigned) cf_sbuf_len) {
+			/*
+			 * We need to handle the complete packet, but it is too
+			 * large to fit into our sbuf buffer size (determined
+			 * by the pkt_buf config). So now we need to fetch the
+			 * whole packet using our dynamically sized packet
+			 * buffering logic.
+			 */
+			client->packet_cb_state.flag = CB_WANT_COMPLETE_PACKET;
+			sbuf_prepare_fetch(sbuf, pkt->len);
+			return true;
+		} else {
+			/*
+			 * We need to handle the complete packet, but it fits
+			 * in our sbuf buffer, so we can simply return false to
+			 * indicate to sbuf to retry once it has received more
+			 * data
+			 */
+			return false;
+		}
+	}
+
+	if (ps_action == PS_INSPECT_FAILED) {
+		if (!incomplete_pkt(pkt)) {
+			/*
+			 * We have the full packet, but still inspection
+			 * failed. That means the packet is plain wrong.
+			 */
+			slog_error(client, "failed to parse prepared statement packet type '%c'", pkt->type);
+			disconnect_client(client, true, "failed to parse packet");
+			return false;
+		}
+
+		/*
+		 * We don't have the full packet yet, so probably inspection
+		 * failed because the required part of the packet was not
+		 * received yet.
+		 */
+
+		if (pkt->data.write_pos >= (unsigned) cf_sbuf_len) {
+			/*
+			 * We've filled up our complete sbuf buffer with this
+			 * packet, but we still haven't been able to determine
+			 * if we should handle this packet or not. This is
+			 * quite unexpected, and probably means that the
+			 * name of the prepared statement is larger than
+			 * pkt_buf.
+			 */
+			client->packet_cb_state.flag = CB_WANT_COMPLETE_PACKET;
+			sbuf_prepare_fetch(sbuf, pkt->len);
+			return true;
+		}
+		/*
+		 * In all other cases we simply return false to indicate to
+		 * sbuf to retry after receiving more data.
+		 */
+		return false;
+	}
+
+	if (ps_action != PS_IGNORE && pkt->type == 'C') {
+		if (!unmarshall_close_packet(client, pkt, &close_packet))
+			return false;
+
+		if (is_close_named_statement_packet(&close_packet)) {
+			if (!handle_close_statement_command(client, pkt, &close_packet))
+				return false;
+
+			client->pool->stats.client_bytes += pkt->len;
+
+			/* No further processing required */
+			return true;
+		}
+	}
+
+
 	/* update stats */
 	if (!client->query_start) {
 		client->pool->stats.query_count++;
@@ -1109,7 +1219,60 @@ static bool handle_client_work(PgSocket *client, PktHdr *pkt)
 	client->link->ready = false;
 	client->link->idle_tx = false;
 
+	if (ps_action != PS_IGNORE) {
+		/*
+		 * All the following handle_xxx_packet functions below insert packets
+		 * into the packet queue through the extra_packets field of SBuf. This
+		 * requires that all previous data in the iobuf is flushed. So lets
+		 * just do that now, so that these functions don't have to worry about
+		 * doing that.
+		 */
+		if (!sbuf_flush(sbuf))
+			return false;
+
+		switch (pkt->type)
+		{
+		case 'P':
+			return handle_parse_command(client, pkt);
+		case 'B':
+			return handle_bind_command(client, pkt);
+		case 'D':
+			return handle_describe_command(client, pkt);
+		}
+		return true;
+	}
+
 	/* forward the packet */
+	if (track_outstanding) {
+		if (!add_outstanding_request(client, pkt->type, RA_FORWARD)) {
+			/* TODO disconnect oom */
+			return false;
+		}
+	}
+
+	if (client->packet_cb_state.flag == CB_HANDLE_COMPLETE_PACKET) {
+		/*
+		 * It's possible that the prepared statement logic required fully
+		 * buffering the packet for inspection purposes using our callback
+		 * packet buffering logic. But once it was fully buffered and the
+		 * inspection caused us to determine that we should simply forward the
+		 * packet (e.g. it was a Describe for a Portal). In those cases we
+		 * cannot simply call sbuf_prepare_send, because we already consumed
+		 * the packet using our callback logic. So now we need to first flush
+		 * the queue, and then re-queue the fully buffered packet using our
+		 * packet queueing logic.
+		 */
+		if (!sbuf_flush(sbuf))
+			return false;
+
+		if (!sbuf_queue_full_packet(&client->sbuf, &client->link->sbuf, pkt)) {
+			disconnect_client(client, true, "out of memory");
+			disconnect_server(client->link, true, "out of memory");
+			return false;
+		}
+		return true;
+	}
+
 	sbuf_prepare_send(sbuf, &client->link->sbuf, pkt->len);
 
 	return true;
@@ -1134,7 +1297,7 @@ bool client_proto(SBuf *sbuf, SBufEvent evtype, struct MBuf *data)
 	switch (evtype) {
 	case SBUF_EV_CONNECT_OK:
 	case SBUF_EV_CONNECT_FAILED:
-		/* ^ those should not happen */
+	/* ^ those should not happen */
 	case SBUF_EV_RECV_FAILED:
 		/*
 		 * Don't log error if client disconnects right away,
@@ -1193,6 +1356,9 @@ bool client_proto(SBuf *sbuf, SBufEvent evtype, struct MBuf *data)
 			break;
 		case CL_WAITING:
 			fatal("why waiting client in client_proto()");
+		case CL_WAITING_CANCEL:
+		case CL_ACTIVE_CANCEL:
+			fatal("why canceling client in client_proto()");
 		default:
 			fatal("bad client state: %d", client->state);
 		}
@@ -1201,8 +1367,71 @@ bool client_proto(SBuf *sbuf, SBufEvent evtype, struct MBuf *data)
 		/* client is not interested in it */
 		break;
 	case SBUF_EV_PKT_CALLBACK:
-		/* unused ATM */
+	{
+		bool first = false;
+		if (client->packet_cb_state.pkt.type == 0) {
+			first = true;
+			if (!get_header(data, &client->packet_cb_state.pkt)) {
+				char hex[8*2 + 1];
+				disconnect_client(client, true, "bad packet header: '%s'",
+						  hdr2hex(data, hex, sizeof(hex)));
+				return false;
+			}
+			mbuf_rewind_reader(data);
+		}
+
+		switch (client->packet_cb_state.flag) {
+		case CB_WANT_COMPLETE_PACKET:
+			if (first) {
+				slog_debug(client,
+					   "buffering complete packet, pkt='%c' len=%d incomplete=%s available=%d",
+					   pkt_desc(&client->packet_cb_state.pkt),
+					   client->packet_cb_state.pkt.len,
+					   incomplete_pkt(&client->packet_cb_state.pkt) ? "true" : "false",
+					   mbuf_avail_for_read(data));
+
+				mbuf_init_dynamic(&client->packet_cb_state.pkt.data);
+				if (!mbuf_make_room(&client->packet_cb_state.pkt.data, client->packet_cb_state.pkt.len))
+					return false;
+			}
+
+			if (!mbuf_write_raw_mbuf(&client->packet_cb_state.pkt.data, data))
+				return false;
+
+			if (sbuf->pkt_remain != mbuf_avail_for_read(data)) {
+				/*
+				 * We wrote the partial packet to our temporary buffer. So
+				 * we "handled" it and want to receive more data.
+				 */
+				res = true;
+				break;
+			}
+
+			/*
+			 * We wrote the full packet into memory. Change the callback state
+			 * to indicate that. If anything fails while handling this packet
+			 * we'll continue from the current state in the callback state
+			 * machine.
+			 */
+			client->packet_cb_state.flag = CB_HANDLE_COMPLETE_PACKET;
+		/* fallthrough */
+		case CB_HANDLE_COMPLETE_PACKET:
+			/* Make sure we start reading after the header. */
+			pkt_rewind_v3(&client->packet_cb_state.pkt);
+			res = handle_client_work(client, &client->packet_cb_state.pkt);
+			if (!res) {
+				return false;
+			}
+
+			client->packet_cb_state.flag = CB_NONE;
+			free_header(&client->packet_cb_state.pkt);
+			break;
+		default:
+			disconnect_client(client, true, "BUG: unknown packet callback flag");
+			break;
+		}
 		break;
+	}
 	case SBUF_EV_TLS_READY:
 		sbuf_continue(&client->sbuf);
 		res = true;
