@@ -107,10 +107,12 @@ def sudo(command, *args, shell=True, **kwargs):
         return run(["sudo", *command])
 
 
+def capture(command, *args, stdout=subprocess.PIPE, encoding="utf-8", **kwargs):
+    return run(command, *args, stdout=stdout, encoding=encoding, **kwargs).stdout
+
+
 def get_pg_major_version():
-    full_version_string = run(
-        "initdb --version", stdout=subprocess.PIPE, encoding="utf-8", silent=True
-    ).stdout
+    full_version_string = capture("initdb --version", silent=True)
     major_version_string = re.search("[0-9]+", full_version_string)
     assert major_version_string is not None
     return int(major_version_string.group(0))
@@ -135,6 +137,10 @@ LONG_PASSWORD = "a" * MAX_PASSWORD_LENGTH
 
 PG_SUPPORTS_SCRAM = PG_MAJOR_VERSION >= 10
 
+# psycopg.Pipeline.is_supported() does not work on rocky:8 in CI, so we create
+# our own check here that works on all our supported systems
+LIBPQ_SUPPORTS_PIPELINING = psycopg.pq.version() >= 140000
+
 
 def get_tls_support():
     with open("../config.mak", encoding="utf-8") as f:
@@ -198,7 +204,10 @@ class QueryRunner:
             options.setdefault("connect_timeout", 20)
         else:
             options.setdefault("connect_timeout", 3)
-        # needed for Ubuntu 18.04
+        # Always required for Ubuntu 18.04, but also needed for any tests
+        # involving the varcache_change database. The difference between the
+        # client_encoding specified in the config and client_encoding by the
+        # client will force a varcache change when a connectino is given.
         options.setdefault("client_encoding", "UTF8")
 
     def conn(self, *, autocommit=True, **kwargs):
@@ -446,6 +455,20 @@ class QueryRunner:
                 elif BSD:
                     sudo(f"pfctl -a pgbouncer_test/port_{self.port} -F all")
 
+    @contextmanager
+    def add_latency(self):
+        """Adds one second of latency to all packets to this query runner"""
+        if not LINUX:
+            raise Exception("This OS cannot run this test")
+        sudo(
+            f"tc filter add dev lo parent 1:0 protocol ip prio {self.port} u32 match ip dport {self.port} 0xffff flowid 1:2"
+        )
+        try:
+            yield
+        finally:
+            sudo(f"tc filter del dev lo parent 1: prio {self.port}")
+            pass
+
 
 class Postgres(QueryRunner):
     def __init__(self, pgdata):
@@ -455,6 +478,7 @@ class Postgres(QueryRunner):
         self.log_path = self.pgdata / "pg.log"
         self.connections = {}
         self.cursors = {}
+        self.restarted = False
 
     def initdb(self):
         run(
@@ -476,6 +500,10 @@ class Postgres(QueryRunner):
             # of our test data.
             pgconf.write("fsync = false\n")
 
+            # Use a consistent value across postgres versions, so test results
+            # are the same.
+            pgconf.write("extra_float_digits = 1\n")
+
     def pgctl(self, command, **kwargs):
         run(f"pg_ctl -w --pgdata {self.pgdata} {command}", **kwargs)
 
@@ -501,6 +529,7 @@ class Postgres(QueryRunner):
         self.port_lock.release()
 
     def restart(self):
+        self.restarted = True
         self.stop()
         self.start()
 
@@ -655,6 +684,8 @@ class Bouncer(QueryRunner):
                 ini.write(f"logfile = {self.log_path}\n")
                 ini.write(f"auth_file = {self.auth_path}\n")
                 ini.write("pidfile = \n")
+                # Uncomment for much more noise but, more detailed debugging
+                # ini.write("verbose = 3\n")
 
                 if not USE_UNIX_SOCKETS:
                     ini.write(f"unix_socket_dir = \n")
@@ -834,3 +865,26 @@ class Bouncer(QueryRunner):
             else:
                 match_count = len(re.findall(re_string, content))
                 assert match_count == times
+
+    @contextmanager
+    def run_with_config(self, config):
+        """Run the pgbouncer instance with provided config and restore the
+        previous config after execution
+
+        config:
+            A new pgbouncer config in ini format
+        """
+        with self.ini_path.open("r") as f:
+            config_old = f.read()
+
+        with self.ini_path.open("w") as f:
+            f.write(config)
+
+        try:
+            self.admin("RELOAD")
+            yield self
+        finally:
+            # Code to release resource, e.g.:
+            with self.ini_path.open("w") as f:
+                f.write(config_old)
+            self.admin("RELOAD")
